# ruby code to download edit histories of articles from wikipedia, read them back in, and some examples of 'automata-like games' you can play with them
# Suggestions, extensions, and thoughts technical or conceptual both welcome and solicited.
# This code is *not debugged* and not for distribution beyond CSSS 2012.

# Simon DeDeo, Santa Fe Institute, 8 June 2012
# simon@santafe.edu

# NOTES
# 1. For speed, we rely quite a bit on regular expressions and bit operations.
# 2. "unit_entropy" will be a basic concept for Jim Crutchfield's lectures (and, in particular, the joint lab with Ryan James')
# 3. How to generate the plot at the end of the lecture:
# download_files("George_W._Bush", 'gwb_edits') # (and wait)
# stream=load_file('gwb_edits')
# Array.new(200) { |i| stream.scan(Regexp.new("#{"C"*(i+1)}")).length }
# note that we plotted the range 10--200; finite size effects (the "tails" of generalized tadpoles) dominate below this threshold.

require 'nokogiri'
require 'time'
require 'open-uri'

class Array # add some useful methods to the Array class
  def nice_a
    self.to_s.gsub(/[^A-Z]/, "")
  end
  
  def total
    self.inject(0) { |num,i|
      num+i      
    }
  end
  
  def mean
    self.inject(0) { |num,i|
      num + i
    }/(1.0*self.length)
  end
  
  def hi(level=0.99)
    self.sort[(self.length-1)*level]
  end
  
  def lo(level=0.99)
    self.sort[(self.length-1)*(1-level)]    
  end
end

def download_files(page_name="September_11_attacks", filename="911_edits") # reads in wikipedia's edit summaries, 5000 units at a time.
  date_type=Regexp.new(/[0-9]{2}:[0-9]{2}, [0-9]{1,2} [A-Z][a-z]{1,} [0-9]{4}/)
  time_type=Regexp.new(/[0-9]{2}:[0-9]{2}/)
  later_type=Regexp.new(/[0-9]{1,2} [A-Z][a-z]{1,} [0-9]{4}/)

  rv_type=Regexp.new(/([Rr]v |[Uu]ndid|[Rr]vv|[Rr]evert)/)

  next_url="http://en.wikipedia.org/w/index.php?title=#{page_name}&offset=&limit=5000&action=history"
  list=[]
  100.times { |i|
    print "Getting #{i} of 100\n"
    start=Time.now
    begin
      doc=Nokogiri::XML(open(next_url).read) # omf omf omf
    rescue Timeout::Error
      print "Ouch! Trying again.\n"
      retry
    else
      print "Got it\n"
      doc.css('li').each { |link| 
        if link.css('span[@class=history-user]').length > 0 then
          current=Hash.new(nil)
          link.css('span[@class=history-user]').each { |sub|
            current[:name] = sub.content.split(" ")[0]
          }
          if date_type.match(link.content) then
            # print "#{Time.parse(time_type.match(link.content).to_a[0]+" UTC "+later_type.match(link.content).to_a[0])}\n"
            current[:time] = Time.parse(time_type.match(link.content).to_a[0]+" UTC "+later_type.match(link.content).to_a[0])
          end
          link.css('span[@class=comment]').each { |sub|
            if rv_type.match(sub.content) then
              current[:rv] = true
            end
          }
          list += [current]
        end
      }
      print "(Took #{(Time.now-start)/60.0} minutes)\n"
      if doc.css('a[@class=mw-nextlink]')[0] == nil then
        break
      else
        next_url="http://en.wikipedia.org"+doc.css('a[@class=mw-nextlink]')[0].values[0]
      end
    end
  }
  file=File.new("#{filename}.dat", 'w')
  file.write(Marshal.dump(list)); 1
  file.close
end

def load_file(filename="911_edits") # reads in downloaded data from a file: drops everything by the coarse-grained edit summary
  file=File.new("#{filename}.dat", 'r')
  list=Marshal.load(file.read)
  file.close
  
  stream=list.collect { |i|
    i[:rv] ? "R" : "C"
  }

  stream.nice_a
end

def pump_test(q=3, text) # checking for FSA of size q

  n_word=0
  n_bad=0
  str_pump=Array.new(2**q) { nil }

  (2**q).times { |i|
    right=i.to_s(2)
    left=(q-right.length)
    str=("0"*left+right).gsub("0", "R").gsub("1", "C")
    # this is the word w we look for.
  
    str_m=Regexp.new(str)
    if str_m.match(text) then # word found
      n_word += 1
      print "#{str} found in stream. Looking for pumped version.\n"

      # we will try all the different ways to pump the string
      best_pump=[]
      0.upto(q-1) { |ystart|
        yend=q
        if ystart > 0 then
          x=str[0..ystart-1]
        else
          x=""
        end
        y=str[ystart..yend]


        best_pump += [[x,y,1.0*text.scan(Regexp.new(x+y*2)).length/text.scan(Regexp.new(x+y)).length,1.0*text.scan(Regexp.new(x+y*3)).length/text.scan(Regexp.new(x+y)).length,1.0*text.scan(Regexp.new(x+y*4)).length/text.scan(Regexp.new(x+y)).length]]
       }
      str_pump[i]=best_pump
      
      # print "For word #{str}, best pump is #{max_pow}, #{best_pump}\n"
    end
  
  }

  return str_pump
end

def CFG_pump_test(q=3, text) # checking for CF of size q

  n_word=0
  n_bad=0
  pump_max=Array.new(2**q) { -1 }
  str_pump=Array.new(2**q) { nil }

  (2**q).times { |i|
    right=i.to_s(2)
    left=(q-right.length)
    str=("0"*left+right).gsub("0", "R").gsub("1", "C")
    # this is the word w we look for.
  
    str_m=Regexp.new(str)
    if str_m.match(text) then # word found
      n_word += 1
      # print "#{str} found in stream. Looking for pumped version.\n"

      # we will try all the different ways to pump the string
      max_pow=1
      best_pump=""
      0.upto(q-3) { |vend|
        (vend+2).upto(q-1) { |ystart|
          v=str[0..vend]
          x=str[vend+1..ystart-1]
          y=str[ystart..q-1]
          
          str_m=Regexp.new(v*(max_pow+1)+x+y*(max_pow+1))
          while (str_m.match(text)) do
            max_pow += 1
            str_m=Regexp.new(v*max_pow+x+y*max_pow)
            best_pump=[v,x,y, max_pow]
           end
          
        }
      }
      if max_pow == 1 then
        n_bad += 1
      end
      pump_max[i]=max_pow
      str_pump[i]=best_pump
      
      # print "For word #{str}, best pump is #{max_pow}, #{best_pump}\n"
    end
  
  }
  # print "#{n_word} found; #{n_bad*100.0/(1.0*n_word)}% unpumped.\n"

  return str_pump
end

def random_string(len)
  
  Array.new(len) {
    rand(2) == 0 ? "R" : "C"
  }.nice_a
end

def random_string_base(len, unit, text) # makes a random string based on the frequencies of length unit in the text
  
  names=[]
  counts=Array.new(2**unit) { |i|
    right=i.to_s(2)
    left=(unit-right.length)
    str=("0"*left+right).gsub("0", "R").gsub("1", "C")
    names += [str]
    str_m=Regexp.new(str)
    text.scan(str_m).length
  }
  freq=counts.collect { |i|
    i*1.0/counts.total
  }
  stair=Array.new(2**unit) { |i|
      freq[0..i].total
  }
  
  Array.new(len/unit) {
    names[stair.index { |i| rand(0) < i }]
  }.nice_a
end

def random_string_sync(len, unit, text) # makes a random string based on the frequencies of length unit in the text
  
  names=[]
  splitter=Regexp.new("[A-Z]{#{unit}}")
  
  list=text.gsub(splitter, "\\0 ").split(" ")
  if (list[-1].length < unit) then
    list=list[0..-2]
  end
  
  found = list & list
  counts=Array.new(2**unit) { 0 }; 1
  names=Array.new(2**unit) { |i|
    right=i.to_s(2)
    left=(unit-right.length)
    ("0"*left+right).gsub("0", "R").gsub("1", "C")    
  }; 1
  
  found.each { |i|
    counts[i.gsub("R", "0").gsub("C", "1").to_i(2)] = list.select { |j| i == j }.length
  }
  freq=counts.collect { |i|
    i*1.0/counts.total
  }
  stair=Array.new(2**unit) { |i|
      freq[0..i].total
  }
  
  Array.new(len/unit) {
    roll=rand(0)
    names[stair.index { |i| roll < i }]
  }.nice_a
end

def unit_entropy(unit, text) # entropy of text blocks

  splitter=Regexp.new("[A-Z]{#{unit}}")  
  list=text.gsub(splitter, "\\0 ").split(" ")
  if (list[-1].length < unit) then
    list=list[0..-2]
  end
  
  found = list & list
  counts=Array.new(2**unit) { 0 }
  
  found.each { |i|
    counts[i.gsub("R", "0").gsub("C", "1").to_i(2)] = list.select { |j| i == j }.length
  }
  freq=counts.collect { |i|
    i*1.0/counts.total
  }
  freq.inject(0) { |num,i|
    num - i*Math::log2(i+1e-64)
  }
end

def string_test(len, text)
  
  nfound=0
  not_found=[]
  (2**len).times { |i|
    right=i.to_s(2)
    left=(len-right.length)
    str=("0"*left+right).gsub("0", "R").gsub("1", "C")
    str_m=Regexp.new(str)
    if str_m.match(text) then
      nfound += 1
    end
  }
  
  [2**len-nfound, nfound*1.0/2**len, text.length*1.0/2**len]
end

def string_test_sync(unit, text)

  splitter=Regexp.new("[A-Z]{#{unit}}")  
  list=text.gsub(splitter, "\\0 ").split(" ")
  if (list[-1].length < unit) then
    list=list[0..-2]
  end

  counts=Array.new(2**unit) { 0 }
  found = list & list
  
  found.each { |i|
    counts[i.gsub("R", "0").gsub("C", "1").to_i(2)] = list.select { |j| i == j }.length
  }
  nfound=counts.select { |i| i > 0 }.length
  
  [2**unit-nfound, nfound*1.0/2**unit, text.length*1.0/2**unit]
end

